use crate::parser::Body;
use crate::parser::DustTag;
use crate::parser::Template;
use crate::parser::{Path, TemplateElement};
use std::collections::HashMap;

pub struct InlinePartialTreeValue<'a> {
    explicit_context: &'a Option<Path<'a>>,
    body: &'a Option<Body<'a>>,
}

pub struct InlinePartialTreeElement<'a> {
    parent: Option<&'a InlinePartialTreeElement<'a>>,
    blocks: HashMap<&'a str, InlinePartialTreeValue<'a>>,
}

impl<'a> InlinePartialTreeElement<'a> {
    pub fn new(
        parent: Option<&'a InlinePartialTreeElement<'a>>,
        blocks: HashMap<&'a str, InlinePartialTreeValue<'a>>,
    ) -> InlinePartialTreeElement<'a> {
        InlinePartialTreeElement {
            parent: parent,
            blocks: blocks,
        }
    }

    fn get_value(&self, name: &str) -> Option<&InlinePartialTreeValue<'a>> {
        match self.blocks.get(name) {
            None => match self.parent {
                None => None,
                Some(parent_tree_element) => parent_tree_element.get_value(name),
            },
            Some(interior) => Some(interior),
        }
    }

    /// Recursively get the body of a block by name
    ///
    /// Returning a double-option because there is a functional
    /// difference between not finding the block and not having a
    /// body.
    pub fn get_block(&self, name: &str) -> Option<&'a Option<Body<'a>>> {
        self.get_value(name).map(|interior| interior.body)
    }

    /// Recursively get the explicit from the inline partial by name
    ///
    /// Returning a double-option because there is a functional
    /// difference between not finding the block and not having a
    /// body.
    pub fn get_explicit_context(&self, name: &str) -> Option<&Option<Path<'a>>> {
        self.get_value(name)
            .map(|interior| interior.explicit_context)
    }
}

pub fn extract_inline_partials<'a>(
    template: &'a Template<'a>,
) -> HashMap<&'a str, InlinePartialTreeValue<'a>> {
    let mut blocks: HashMap<&'a str, InlinePartialTreeValue<'a>> = HashMap::new();

    extract_inline_partials_from_body(&mut blocks, &template.contents);

    blocks
}

fn extract_inline_partials_from_body<'a, 'b>(
    blocks: &'b mut HashMap<&'a str, InlinePartialTreeValue<'a>>,
    body: &'a Body<'a>,
) {
    for elem in &body.elements {
        match elem {
            TemplateElement::TEIgnoredWhitespace(_) => (),
            TemplateElement::TESpan(span) => (),
            TemplateElement::TETag(dt) => {
                extract_inline_partials_from_tag(blocks, dt);
            }
        }
    }
}

fn extract_inline_partials_from_tag<'a, 'b>(
    blocks: &'b mut HashMap<&'a str, InlinePartialTreeValue<'a>>,
    tag: &'a DustTag,
) {
    match tag {
        DustTag::DTComment(..) => (),
        DustTag::DTSpecial(..) => (),
        DustTag::DTLiteralStringBlock(..) => (),
        DustTag::DTReference(..) => (),
        DustTag::DTSection(container) => {
            match &container.contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
            match &container.else_contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
        }
        DustTag::DTExists(container) => {
            match &container.contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
            match &container.else_contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
        }
        DustTag::DTNotExists(container) => {
            match &container.contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
            match &container.else_contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
        }
        DustTag::DTPartial(..) => (),
        DustTag::DTInlinePartial(named_block) => {
            blocks.insert(
                &named_block.name,
                InlinePartialTreeValue {
                    explicit_context: &named_block.explicit_context,
                    body: &named_block.contents,
                },
            );
        }
        DustTag::DTBlock(..) => (),
        DustTag::DTHelperEquals(parameterized_block) => {
            match &parameterized_block.contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
            match &parameterized_block.else_contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
        }
        DustTag::DTHelperNotEquals(parameterized_block) => {
            match &parameterized_block.contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
            match &parameterized_block.else_contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
        }
        DustTag::DTHelperGreaterThan(parameterized_block) => {
            match &parameterized_block.contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
            match &parameterized_block.else_contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
        }
        DustTag::DTHelperLessThan(parameterized_block) => {
            match &parameterized_block.contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
            match &parameterized_block.else_contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
        }
        DustTag::DTHelperGreaterThanOrEquals(parameterized_block) => {
            match &parameterized_block.contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
            match &parameterized_block.else_contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
        }
        DustTag::DTHelperLessThanOrEquals(parameterized_block) => {
            match &parameterized_block.contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
            match &parameterized_block.else_contents {
                None => (),
                Some(body) => extract_inline_partials_from_body(blocks, &body),
            };
        }
    }
}
